#ifndef	_MY_SCSI_H_
#define	_MY_SCSI_H_

/* in addtion to drivers/Drivers.h */
enum
{
	B_GRAPHIC_A = 0x0a,
	B_GRAPHIC_B,
	B_RAID,
	B_ENCLOSURE,
	B_RBC_DISK,
	B_CARD_READER,
	B_BRIDGE,
	B_OSD,
	B_RMC_CD = 0x15,
};

/* in addition to device/scsi.h */
enum
{
	B_SCSI_GRAPHIC_A = 0x0a,
	B_SCSI_GRAPHIC_B,
	B_SCSI_RAID,
	B_SCSI_ENCLOSURE,
	B_SCSI_RBC_DISK,
	B_SCSI_CARD_READER,
	B_SCSI_BRIDGE,
	B_SCSI_OSD,
	B_SCSI_RMC_CD = 0x15,
};

/* SCSI command ops */

enum
{
	/* 6-byte commands */
	TEST_UNIT_READY = 0x00,
	REZERO_UNIT = 0x01,
	REQUEST_SENSE = 0x03,
	FORMAT_UNIT = 0x04,
	INQUIRY = 0x12, 
	START_STOP_UNIT = 0x1b,
	SEND_DIAGNOSTIC = 0x1d,
	PREVENT_ALLOW_MEDIUM_REMOVAL = 0x1e,

	/* 10-byte commands */
	READ_FORMAT_CAPACITIES = 0x23,	/* UFI specific */
	READ_CAPACITY = 0x25,
	READ_10 = 0x28,
	WRITE_10 = 0x2a,
	SEEK_10 = 0x2b,
	WRITE_AND_VERIFY_10 = 0x2e,
	VERIFY_10 = 0x2f,
	SYNCHRONIZE_CACHE = 0x35,

	READ_SUB_CHANNEL = 0x42,
	READ_TOC = 0x43,
	PLAY_AUDIO_MSF = 0x47,
	PLAY_AUDIO_TRACK_INDEX = 0x48,
	PAUSE_RESUME = 0x4b,
	STOP_PLAY_SCAN = 0x4e,
	MODE_SELECT_10 = 0x55,
	MODE_SENSE_10 = 0x5a,

	/* 12-byte commands */

	READ_12 = 0xa8,
	WRITE_12 = 0xaa,
	READ_CD_MSF = 0xb9,
	SCAN = 0xba,
};

/* parameters for CD-DA functions */
enum
{
	MODE_LBA = 0,
	MODE_MSF,
	MODE_TNO
};

enum
{
	DIR_FWD = 0,
	DIR_REW
};

typedef struct std_inquiry_data
{
	/* offset 0 */
	uint8	device_type:5;
	uint8	qualifier:3;
	/* offset 1 */
	uint8	device_type_modifier:7;
	uint8	removable:1;
	/* offset 2 */
	uint8	ansi_version:3;
	uint8	ecma_version:3;
	uint8	iso_version:2;
	/* offset 3 */
	uint8	response_data_format:4;
	uint8	reserved_3:2;
	uint8	trmiop:1;
	uint8	aenc:1;
	/* offset 4-6 */
	uint8	additional_length;
	uint8	reserved_5;
	uint8	reserved_6;
	/* offset 7 */
	uint8	sftre:1;
	uint8	cmdque:1;
	uint8	reserved_7:1;
	uint8	linked:1;
	uint8	sync:1;
	uint8	wbus16:1;
	uint8	wbus32:1;
	uint8	reladr:1;
	/* offset 8- */
	uchar	vendor_id [8];
	uchar	product_id [16];
	uchar	product_version [4];
#if 1
	uint8	vendor_unique_1 [20];
	uint8	reserved_56 [40];
	uint8	vendor_unique_2 [0];
#endif
} __attribute__((packed)) std_inquiry_data;

/* CCS/SCSI-2 extended format */
typedef struct sense_data
{
	/* offset 0 */
	uint8	error_code:7;	/* 0x70 or 0x71 */
	uint8	valid:1;
	/* offset 1 */
	uint8	segment_num;
	/* offset 2 */
	uint8	sense_key:4;
	uint8	reserved_2:1;
	uint8	ili:1;
	uint8	eom:1;
	uint8	filemark:1;
	/* offset 3-6 */
	uint8	info [4];
	/* offset 7 */
	uint8	additional_length;
	/* offset 8-11 */
	uint8	cmd_specific [4];
	/* offset 12-14 */
	uint8	asc, ascq, fru_code;
	/* offset 15 */
	uint8	sense_key_specific_msb:7;
	uint8	sksv:1;
	/* offset 16-17 */
	uint8	sense_key_specific [2];
	/* offset 18- */
	uint8	additional_bytes [0];
} __attribute__((packed)) sense_data;

typedef struct mode_param_header_4
{
	uint8	mode_param_length;
	uint8	medium_type;
	uint8	device_specific;
	uint8	block_desc_length;
} __attribute__((packed)) mode_param_header_4;

typedef struct mode_param_header_8
{
	uint8	mode_param_length [2];
	uint8	medium_type;

	uint8	reserved_3a:4;
	uint8	dpofua:1;
	uint8	reserved_3b:2;
	uint8	wp:1;

	uint8	reserved [2];
	uint8	block_desc_length [2];
} __attribute__((packed)) mode_param_header_8;

typedef struct block_desc
{
	uint8	density_code;
	uint8	num_blocks [3];
	uint8	reserved;
	uint8	block_length [3];
} __attribute__((packed)) block_desc;

/* page 01h */
typedef struct read_write_error_recovery_page
{
	/* offset 0 */
	uint8	page_code:6;
	uint8	reserved_0:1;
	uint8	ps:1;
	/* offset 1 */
	uint8	length;		/* 10 */
	/* offset 2 */
	uint8	dcr:1;
	uint8	dte:1;
	uint8	per:1;
	uint8	eer:1;
	uint8	rc:1;
	uint8	tb:1;
	uint8	arre:1;
	uint8	awre:1;
	/* offset 3- */
	uint8	read_retry_count;
	uint8	data_correction_length;
	uint8	head_offset;
	uint8	data_strobe_offset;
	uint8	reserved_7;
	uint8	write_retry_count;
	uint8	reserved_9;
	uint8	recovery_time_max [2];
} __attribute__((packed)) read_write_error_recovery_page;

/* page 03h */
typedef struct format_page
{
} format_page;

/* page 04h */
typedef struct fixed_drive_page
{
	/* offset 0 */
	uint8	page_code:6;
	uint8	reserved_0:1;
	uint8	ps:1;
	/* offset 1 */
	uint8	length;		/* 22 */
	/* offset 2-16 */
	uint8	num_cylinders [3];
	uint8	num_heads;
	uint8	write_precomp_cylinder [3];
	uint8	write_low_current_cylinder [3];
	uint8	drive_step_rate [2];
	uint8	landing_zone_cylinder [3];
	/* offset 17 */
	uint8	rpl:2;
	uint8	reserved_17:6;
	/* offset 18- */
	uint8	rot_sync_offset;
	uint8	reserved_19;
	uint8	rot_rate [2];
	uint8	reserved_22;
	uint8	reserved_23;
} __attribute__((packed)) fixed_drive_page;

/* page 05h */
typedef struct flexible_disk_page
{
	/* offset 0 */
	uint8	page_code:6;
	uint8	reserved_0:1;
	uint8	ps:1;
	/* offset 1 */
	uint8	length;		/* 30 */
	/* offset 2-9 */
	uint8	xfer_rate [2];
	uint8	num_heads;
	uint8	num_sectors_per_track;
	uint8	sector_size [2];
	uint8	num_cylinders [2];
	/* offset 10-18 */
	uint8	write_precomp_cylinder [2];
	uint8	write_low_current_cylinder [2];
	uint8	drive_step_rate [2];
	uint8	drive_step_pulse;
	uint8	head_settling_delay [2];
	/* offset 19-20 */
	uint8	motor_on_delay;
	uint8	motor_off_delay;
	/* offset 21 */
	uint8	reserved_21:5;
	uint8	mo:1;
	uint8	ssn:1;
	uint8	trdy:1;
	/* offset 22 */
	uint8	spc:4;
	uint8	reserved_22:4;
	/* offset 23-24 */
	uint8	write_precomp;
	uint8	head_load_delay;
	uint8	head_unload_delay;
	/* offset 25 */
	uint8	pin2:4;
	uint8	pin34:4;
	/* offset 26 */
	uint8	pin1:4;
	/* offset 27 */
	uint8	pin4:4;
	/* offset 28-31 */
	uint8	rot_rate [2];
	uint8	reserved_30;
	uint8	reserved_31;
} __attribute__((packed)) flexible_disk_page;

typedef struct audio_ctrl_page
{
	/* offset 0 */
	uint8	page_code:6;
	uint8	reserved_0:1;
	uint8	ps:1;
	/* offset 1 */
	uint8	length;		/* 14 */
	/* offset 2 */
	uint8	reserved_2:1;
	uint8	sotc:1;
	uint8	immed:1;
	uint8	reserved_2a:5;

	uint8	reserved_3;
	uint8	reserved_4;

	uint8	unit:4;
	uint8	reserved_5:3;
	uint8	aprval:1;

	uint8	blocks_per_sec [2];

	struct
	{
		uint8	channels:4;
		uint8	reserved:4;
		uint8	volume;
	} __attribute__((packed)) ports [4];
} __attribute__((packed)) audio_ctrl_page;

typedef struct read_capacity_data
{
	uint8	last_lba [4];
	uint8	block_length [4];
} __attribute__((packed)) read_capacity_data;

/* big-endian read/write */

#define	read4(a) \
	(((a)[0] << 24) | ((a)[1] << 16) | ((a)[2] <<  8) | ((a)[3]))
#define	read3(a) \
	(                 ((a)[0] << 16) | ((a)[1] <<  8) | ((a)[2]))
#define	read2(a) \
	(                                  ((a)[0] <<  8) | ((a)[1]))

#define write4(a,d) \
	((a)[0] = (d) >> 24, (a)[1] = (d) >> 16, (a)[2] = (d) >>  8, (a)[3] = (d))
#define write3(a,d) \
	(                    (a)[0] = (d) >> 16, (a)[1] = (d) >>  8, (a)[2] = (d))
#define write2(a,d) \
	(                                        (a)[0] = (d) >>  8, (a)[1] = (d))

#endif
